package be.iminds.ilabt.jfed.experiment;

import be.iminds.ilabt.jfed.experiment.events.*;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.experiment.util.NextExperimentExpiration;
import be.iminds.ilabt.jfed.experiment.util.NextExperimentExpirationBinding;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Service;
import be.iminds.ilabt.jfed.highlevel.jobs.Job;
import be.iminds.ilabt.jfed.highlevel.jobs.JobFactory;
import be.iminds.ilabt.jfed.highlevel.model.InternalState;
import be.iminds.ilabt.jfed.highlevel.model.Sliver;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.TestbedInfoSource;
import be.iminds.ilabt.jfed.lowlevel.api.user_spec.UserSpec;
import be.iminds.ilabt.jfed.preferences.JFedCorePreferences;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecLink;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXRspecNode;
import be.iminds.ilabt.jfed.util.FXPlatformUtil;
import be.iminds.ilabt.jfed.util.GeniUrn;
import com.google.common.util.concurrent.*;
import javafx.application.Platform;
import javafx.stage.Window;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.*;

/**
 * This class will take care of an Experiment-object. It makes sure that the appropriate actions are taken when
 * the state of the Slice, or one of it's slivers changes.
 * <p/>
 * User: twalcari
 * Date: 1/30/2015
 * Time: 14:04
 */
public class ExperimentController implements ExperimentChangeListener {
    private static final Logger LOG = LoggerFactory.getLogger(ExperimentController.class);
    @Nonnull
    private final Experiment experiment;
    private final ExperimentEventHandlerManager eventHandlerManager = new ExperimentEventHandlerManager();

    private final JobFactory jobFactory;
    private final TestbedInfoSource testbedInfoSource;
    private final JFedCorePreferences jFedPreferences;
    private final GeniUserProvider geniUserProvider;

    private static final ExecutorService executorService = Executors.newCachedThreadPool(
            new ThreadFactoryBuilder().setNameFormat("ExperimentController-%d").build()
    );
    private final ScheduledExecutorService scheduledExecutorService = Executors.newScheduledThreadPool(2,
            new ThreadFactoryBuilder().setNameFormat("ExperimentController-Scheduled-%d").build());


    private static final ListeningExecutorService jobExecutorService =
            MoreExecutors.listeningDecorator(Executors.newCachedThreadPool(
                    new ThreadFactoryBuilder().setNameFormat("JobExecutor-%d").build()));

    private Window parentWindow = null;
    //helper variable for showing dialog messages. Ideally, this should be refactored away

    private final List<ExperimentControllerListener> listeners = new ArrayList<>();
    private final ExpirationDetector expirationDetector;

    private ExperimentTester experimentTester = null;

    public ExperimentController(@Nonnull Experiment experiment,
                                @Nonnull JobFactory jobFactory,
                                @Nonnull TestbedInfoSource testbedInfoSource,
                                @Nonnull JFedCorePreferences jFedPreferences,
                                @Nonnull GeniUserProvider geniUserProvider) {
        this.experiment = experiment;
        this.jobFactory = jobFactory;
        this.testbedInfoSource = testbedInfoSource;
        this.jFedPreferences = jFedPreferences;
        this.geniUserProvider = geniUserProvider;

        this.expirationDetector = new ExpirationDetector();

        experiment.addExperimentChangeListener(this);
    }

    /**
     * This method must be called for the experimentController to start taking actions.
     */
    public void start() {
        //bootstrap state machine
        LOG.debug("Bootstrapping experiment controller of '{}' in state {}", experiment.getName(), experiment.getExperimentState());
        onExperimentStateChange(experiment.getExperimentState());
    }


    @Override
    public void onExperimentStateChange(@Nonnull ExperimentState newExperimentState) {
        switch (newExperimentState) {
            case PENDING:
                createSlice();
                //next: CREATED
                break;
            case RESTORING:
                restoreSlice();
                //next: RESTORED or RENEW_EXISTING  (depending on if it's a recover or a new experiment creation)
                break;
            case RENEW_EXISTING:
                renewExistingSlice();
                //next: CREATED
                break;
            case RESTORED:
                handleRestored();
                //next: FUTURE_RESERVATION, WAIT_FOR_READY, TESTING or FAILED
                break;
            case CREATED:
                //startSliceExpirationDetection();
                assert experiment.getNewRequestRspecSource() != null;
                if (experiment.getRequestedStartTime() != null) {
                    if (!experiment.isReservationMade()) {
                        experiment.setExperimentState(ExperimentState.ALLOCATING);
                    } else if (experiment.getRequestedStartTime().isAfter(Instant.now())) {
                        experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
                    } else {
                        // the reservation allocation was already done, and the start time has passed, so start provisioning
                        experiment.setExperimentState(ExperimentState.PROVISIONING);
                    }
                } else {
                    experiment.setExperimentState(ExperimentState.ALLOCATING);
                }
                break;
            case ALLOCATING:
                allocateResources();
                //next: FUTURE_RESERVATION, PROVISIONING or FAILING
                break;
            case FUTURE_RESERVATION:
                waitForReservationStart();
                //next: PROVISIONING  (after delay)
                break;
            case PROVISIONING:
                provisionResources();
                //next: WAIT_FOR_READY or FAILING
                break;
            case WAIT_FOR_READY:
                waitForReady();
                //next: TESTING or FAILING
                break;
            case TESTING:
                testSliversConnectivity();
                //next: SETUP_SOFTWARE (if create), READY (if recover) or FAILING
                break;
            case SETUP_SOFTWARE:
                setupSoftware();
                //next: READY or FAILING
                break;
            case READY:
                //final state (until user takes an action, or expiration)
                break;
            case FAILING:
                //failure, and currently asking user what to do
                //after user decision, goes to FAILING
                break;
            case FAILED:
                //final state
                break;
            case EXPIRING:
                checkForExpiration();
                break;
            case EXPIRED:
            case EMPTY:
                deleteRestoreInformation();
                break;
        }


    }

    private void deleteRestoreInformation() {
        experiment.getExperimentRestoreInformation().delete();
    }

    @Override
    public void onExperimentPartAdded(ExperimentPart experimentPart) {
        //do nothing
    }

    private void checkForExpiration() {
        requestUpdate();
    }

    private void checkExperimentPartsStates() {
        boolean allPartsDeletedOrExpired = experiment.getPartsListCopy().stream()
                .allMatch(ep -> ep.getState().equals(InternalState.DELETED) || ep.getState().equals(InternalState.UNALLOCATED));

        if (allPartsDeletedOrExpired) {
            if (experiment.getSliceOrNull() != null
                    && experiment.getSliceOrNull().getExpirationDate() != null
                    && experiment.getSliceOrNull().getExpirationDate().isBefore(Instant.now()))
                experiment.setExperimentState(ExperimentState.EXPIRED);
            else
                experiment.setExperimentState(ExperimentState.EMPTY);
        } else {
            boolean hasAnyFailingSliver = experiment.getSliceOrNull() != null &&
                    experiment.getSliceOrNull().getSliversStream()
                            .map(Sliver::hasAnyFailStatus).anyMatch(s -> s);
            if (hasAnyFailingSliver)
                experiment.setExperimentState(ExperimentState.FAILING);
            else
                experiment.setExperimentState(ExperimentState.READY);
        }
    }


    private void testSliversConnectivity() {
        if (experimentTester != null) {
            Future<Boolean> future = submitJob(experimentTester.createTestExperimentJob(experiment));

            Thread testWaiterThread = new Thread(null, () -> {
                try {
                    future.get();
                } catch (InterruptedException | ExecutionException e) {
                    LOG.error("Error while waiting for ExperimentTester of '{}' to finish: {}",
                            experiment.getName(), e.getMessage(), e);
                }

                deriveAndSetExperimentStateAfterTesting();

            }, "Wait-For-Tester-" + experiment.getName());
            testWaiterThread.start();
        } else {
            LOG.info("Skipping testing of experiment, as no ExperimentTester is registered.");
            deriveAndSetExperimentStateAfterTesting();
        }
    }

    private void deriveAndSetExperimentStateAfterTesting() {
        assert experiment.getExperimentState() == ExperimentState.TESTING :
                "experiment.getExperimentState() is not TESTING but " + experiment.getExperimentState();

        NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment);
        if (nextExperimentExpiration.hasPartExpired()) {
            experiment.setExperimentState(ExperimentState.EXPIRING);
            return;
        }

        boolean allReady = true;
        boolean allEmpty = true;

        for (ExperimentPart part : experiment.getPartsListCopy()) {
            LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting part.getState()=" + part.getState());
            allReady = allReady && part.getState() == InternalState.READY;
            allEmpty = allEmpty && part.getState() == InternalState.UNALLOCATED;
        }

        LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting allReady=" + allReady + " allEmpty=" + allEmpty);

        if (allReady) {
            if (experiment.getNewRequestRspecSource() != null) {
                experiment.setExperimentState(ExperimentState.SETUP_SOFTWARE);
            } else {
                experiment.setExperimentState(ExperimentState.READY);
            }
        } else {
            if (allEmpty) {
                experiment.setExperimentState(ExperimentState.EMPTY);
            } else {
                LOG.warn("Could not determine a valid experiment state for {}. Defaulting to allReady", experiment.getName());

                if (experiment.getNewRequestRspecSource() != null) {
                    experiment.setExperimentState(ExperimentState.SETUP_SOFTWARE);
                } else {
                    experiment.setExperimentState(ExperimentState.READY);
                }
            }
        }

        LOG.debug("ExperimentController deriveAndSetExperimentStateAfterTesting post state = " + experiment.getExperimentState());
    }


    public void createSlice() {
        //we must initiate the creation of the slice for this experiment
        assert (experiment.getRequestedEndTime() != null);

        submitJob(jobFactory.createRegisterExperimentJob(experiment));
    }

    private void restoreSlice() {
        submitJob(jobFactory.createRestoreExperimentJob(experiment));
    }

    private void renewExistingSlice() {
        if (experiment.getRequestedEndTime() == null) {
            experiment.setExperimentState(ExperimentState.CREATED);
            return;
        }

        NextExperimentExpiration nextExperimentExpiration = new NextExperimentExpiration(experiment);
        if (nextExperimentExpiration.getFirstExpirationTime() == null ||
                nextExperimentExpiration.getFirstExpirationTime().isBefore(experiment.getRequestedEndTime())) {
            //do renew

            ListenableFuture<ExperimentTaskStatus> future =
                    submitJob(jobFactory.createRenewSliceJob(experiment, experiment.getRequestedEndTime()));

            Futures.addCallback(future, new FutureCallback<ExperimentTaskStatus>() {
                @Override
                public void onSuccess(@Nullable ExperimentTaskStatus renewStatus) {
                    if (renewStatus == ExperimentTaskStatus.SUCCESS) {
                        experiment.setExperimentState(ExperimentState.CREATED);
                    } else {
                        LOG.debug("Renew Slice for {} job success but renew failed: returned renewStatus={}",
                                experiment.getName(), renewStatus);
                        experiment.setExperimentState(ExperimentState.FAILING);
                    }
                }

                @Override
                public void onFailure(@Nonnull Throwable throwable) {
                    LOG.error("Renew Slice for " + experiment.getName() + " failed: ", throwable);
                    experiment.setExperimentState(ExperimentState.FAILING);
                }
            }, FXPlatformUtil.getExecutor());
        } else {
            experiment.setExperimentState(ExperimentState.CREATED);
        }
    }

    private void handleRestored() {
        if (experiment.getRequestedStartTime() != null
                && experiment.getRequestedStartTime().isAfter(Instant.now())) {
            experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
        } else {
            boolean hasFailure = experiment.getPartsStream()
                    .anyMatch(part -> part.getState() == InternalState.FAILED);

            boolean hasChanging = experiment.getPartsStream()
                    .filter(Objects::nonNull)
                    .map(ExperimentPart::getState)
                    .filter(Objects::nonNull)
                    .anyMatch(partState -> {
                        switch (partState) {
                            case CHANGING:
                            case ALLOCATING:
                            case PROVISIONING:
                            case DELETING:
                            case RESERVING:
                            case UNKNOWN:
                            case WAIT_FOR_READY:
                                return true;
                            default:
                                return false;
                        }
                    });

            if (hasFailure) {
                //not FAILING: we don't need to ask user
                experiment.setExperimentState(ExperimentState.FAILED);
                return;
            }

            if (hasChanging) {
                experiment.setExperimentState(ExperimentState.WAIT_FOR_READY);
                return;
            }

            experiment.setExperimentState(ExperimentState.TESTING);
        }
    }

    private void allocateResources() {
        if (experiment.getSliceOrNull() != null) {
            experiment.getSlice().setRequestRspec(experiment.getNewRequestRspecSource());
        }

        if (experiment.getExperimentSpecification() != null) {
            experiment.getExperimentSpecificationLogger().firePreRSpec();
        }

        ListenableFuture<Boolean> future =
                submitJob(jobFactory.createAllocateExperimentJob(experiment, getScsAuthority()));

        Futures.addCallback(future, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(@Nullable Boolean success) {
                if (Objects.equals(success, Boolean.TRUE)) {
                    if (experiment.getRequestedStartTime() == null || experiment.getRequestedStartTime().isBefore(Instant.now())) {
                        experiment.setExperimentState(ExperimentState.PROVISIONING);
                    } else {
                        experiment.setExperimentState(ExperimentState.FUTURE_RESERVATION);
                    }
                } else {
                    LOG.debug("AllocateExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                    experiment.setExperimentState(ExperimentState.FAILING);
                }
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.error("AllocateExperimentJob for " + experiment.getName() + " failed: ", throwable);
                experiment.setExperimentState(ExperimentState.FAILING);
            }
        }, FXPlatformUtil.getExecutor());
    }


    private void waitForReservationStart() {
        assert (experiment.getRequestedStartTime() != null);

        Duration durationToStartTime = Duration.between(Instant.now(), experiment.getRequestedStartTime());

        Runnable startProvisioning = () -> experiment.setExperimentState(ExperimentState.PROVISIONING);

        if (!durationToStartTime.isNegative()) {


            scheduledExecutorService.schedule(startProvisioning, durationToStartTime.getSeconds() + 1, TimeUnit.SECONDS);
        } else {
            startProvisioning.run();
        }
    }

    public ListenableFuture<Void> requestUpdate() {
        final ListenableFuture<Void> future = submitJob(jobFactory.createUpdateExperimentJob(this.getExperiment()));
        Futures.addCallback(future, new FutureCallback<Void>() {
            @Override
            public void onSuccess(@Nullable Void aVoid) {
                checkExperimentPartsStates();
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.warn("Update experiment failed: {}. Checking for experiment part states anyway.", throwable.getMessage(), throwable);
                checkExperimentPartsStates();
            }
        }, Platform::runLater);
        return future;
    }

    public void shareWithUsers(@Nonnull List<GeniUrn> selectedUsers, boolean registerSshKeys) {
        submitJob(jobFactory.createShareSliceJob(experiment, selectedUsers, registerSshKeys));
    }
    public void unshareWithUsers(@Nonnull List<GeniUrn> selectedUsers, boolean unregisterSshKeys) {
        submitJob(jobFactory.createUnshareSliceJob(experiment, selectedUsers, unregisterSshKeys));
    }

    public void updateSshKeys(@Nonnull List<UserSpec> userspec) {
        submitJob(jobFactory.createEditSshKeysJob(experiment, userspec));
    }

    public void reloadOS(Sliver sliver) {
        submitJob(jobFactory.createReloadOSJob(experiment, sliver));
    }

    public void createDiskImage(FXRspecNode node, String imageName, boolean global, boolean updatePrepare) {
        submitJob(jobFactory.createCreateDiskImageJob(experiment, node, imageName, global, updatePrepare));
    }

    public void stop(Collection<ExperimentPart> experimentParts) {
        final ListenableFuture<Void> future =
                submitJob(jobFactory.createStopExperimentJob(experiment, experimentParts));

        Futures.addCallback(future, new FutureCallback<Void>() {
            @Override
            public void onSuccess(@Nullable Void aVoid) {
                checkExperimentPartsStates();
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.warn("Stopping experiment failed: {}. Checking experiment part states anyway.",
                        throwable.getMessage(), throwable);
                checkExperimentPartsStates();
            }
        });

        eventHandlerManager.dispatch(new ExperimentPartsEvent(this, ExperimentPartsEvent.STOPPED, experimentParts));
    }

    public void stop() {
        stop(experiment.getPartsListCopy());
    }

    public void reboot(Sliver sliver) {
        submitJob(jobFactory.createRebootJob(experiment, sliver));
    }

    public void reboot(ExperimentPart experimentPart) {
        assert (experiment.getPartsListCopy().contains(experimentPart));
        submitJob(jobFactory.createRebootJob(experimentPart));
    }

    public void openConsole(Sliver sliver) {
        submitJob(jobFactory.createOpenConsoleJob(experiment, sliver));
    }

    public void shareLan(FXRspecLink link, String sharedLanName) {
        submitJob(jobFactory.createShareLanJob(experiment, link, sharedLanName));
    }

    public void unshareLan(FXRspecLink link, String sharedLanName) {
        submitJob(jobFactory.createUnshareLanJob(experiment, link, sharedLanName));
    }

    public ListenableFuture<Collection<UserSpec>> fetchSliceMemberSshKeys() {
        return submitJob(jobFactory.createFetchSliceMemberSshKeysJob(experiment));
    }

    public ListenableFuture<ExperimentTaskStatus> renew(Instant newExpirationTime) {
        return submitJob(jobFactory.createRenewExperimentJob(experiment, newExpirationTime));
    }

    public void renewParts(Instant newExpirationTime, Collection<ExperimentPart> parts) {
        submitJob(jobFactory.createRenewExperimentJob(experiment, newExpirationTime, parts));
    }


    private class ExpirationDetector implements Runnable {

        private final NextExperimentExpirationBinding nextExperimentExpirationBinding;
        private ScheduledFuture<?> previousScheduledExpiration = null;

        public ExpirationDetector() {
            this.nextExperimentExpirationBinding = new NextExperimentExpirationBinding(experiment);
            this.nextExperimentExpirationBinding.addListener(observable -> scheduleForNextExpiration());
        }


        @Override
        public void run() {
            final NextExperimentExpiration nextExpiration = nextExperimentExpirationBinding.get();
            if (nextExpiration.hasPartExpired()) {
                LOG.trace("A part of the experiment {} has expired!", experiment.getName());
                //(a part of) the experiment seems to have expired!
                experiment.setExperimentState(ExperimentState.EXPIRING);
            } else {
                scheduleForNextExpiration();
            }
        }

        private void scheduleForNextExpiration() {
            if (previousScheduledExpiration != null)
                previousScheduledExpiration.cancel(false);

            final NextExperimentExpiration nextExpiration = nextExperimentExpirationBinding.get();

            if (nextExpiration.getFirstExpirationTime() != null) {
                Duration durationToFirstExpiration =
                        Duration.between(Instant.now(), nextExpiration.getFirstExpirationTime());

                assert !durationToFirstExpiration.isNegative();

                LOG.debug("Scheduling next expiration check for {} in {} seconds", experiment.getName(), durationToFirstExpiration.getSeconds() + 1);

                previousScheduledExpiration =
                        scheduledExecutorService.schedule(this, durationToFirstExpiration.getSeconds() + 1, TimeUnit.SECONDS);
            } else if (!nextExpiration.hasPartExpired()) {
                LOG.debug("Could not find an expiration time. Trying again in 10 seconds");
                previousScheduledExpiration = scheduledExecutorService.schedule(this, 10, TimeUnit.SECONDS);
            }
        }
    }

    /**
     * This will also wait until the resource is ready
     */
    private void provisionResources() {
        assert experiment.getNewRequestRspecSource() != null;
        assert experiment.getSliceOrNull() != null;

        if (experiment.getSliceOrNull() != null) {
            experiment.getSliceOrNull().setRequestRspec(experiment.getNewRequestRspecSource());
        }

        ListenableFuture<Boolean> future =
                submitJob(jobFactory.createProvisionExperimentJob(experiment));

        Futures.addCallback(future, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(@Nullable Boolean success) {
                if (Objects.equals(success, Boolean.TRUE))
                    experiment.setExperimentState(ExperimentState.WAIT_FOR_READY);
                else {
                    LOG.debug("ProvisionExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                    experiment.setExperimentState(ExperimentState.FAILING);
                }
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.error("ProvisionExperimentJob for " + experiment.getName() + " failed: ", throwable);
                experiment.setExperimentState(ExperimentState.FAILING);
            }
        }, FXPlatformUtil.getExecutor());

    }

    public void waitForReady() {
        ListenableFuture<Boolean> future =
                submitJob(jobFactory.createWaitForReadyExperimentJob(experiment));

        Futures.addCallback(future, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(@Nullable Boolean success) {
                if (Objects.equals(success, Boolean.TRUE))
                    experiment.setExperimentState(ExperimentState.TESTING);
                else {
                    LOG.debug("WaitForOpStatusExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                    experiment.setExperimentState(ExperimentState.FAILING);
                }
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.error("WaitForOpStatusExperimentJob for " + experiment.getName() + " failed: ", throwable);
                experiment.setExperimentState(ExperimentState.FAILING);
            }
        }, FXPlatformUtil.getExecutor());
    }

    public void setupSoftware() {
        ListenableFuture<Boolean> future =
                submitJob(jobFactory.createSetupSoftwareExperimentJob(experiment));

        Futures.addCallback(future, new FutureCallback<Boolean>() {
            @Override
            public void onSuccess(@Nullable Boolean success) {
                if (Objects.equals(success, Boolean.TRUE))
                    experiment.setExperimentState(ExperimentState.READY);
                else {
                    LOG.debug("SetupSoftwareExperimentJob for {} succeeded but returned success={}", experiment.getName(), success);
                    experiment.setExperimentState(ExperimentState.FAILING);
                }
            }

            @Override
            public void onFailure(@Nonnull Throwable throwable) {
                LOG.error("SetupSoftwareExperimentJob for " + experiment.getName() + " failed: ", throwable);
                experiment.setExperimentState(ExperimentState.FAILING);
            }
        }, FXPlatformUtil.getExecutor());
    }

    private Server getScsAuthority() {
        Server scsServer = jFedPreferences.getScsAuthority(testbedInfoSource);
        if (scsServer == null) {
            Integer scsId = geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer() != null ?
                    geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer().getDefaultScsId() :
                    null;
            if (scsId != null) {
                Service scsService = testbedInfoSource.getServiceById(scsId);
                assert scsService != null : "Could not find scs " + scsId;
                if (scsService != null) {
                    scsServer = scsService.getServer();
                    assert scsServer != null;
//                  GeniUrn defaultScsUrn = scsServer.getDefaultComponentManagerAsGeniUrn();
                }
            }
        }
        return scsServer;
    }

    private <V> ListenableFuture<V> submitJob(Job<V> job) {

        listeners.forEach(listener -> listener.onJobSubmitted(job));

        job.stateProperty().addListener(observable ->
                listeners.forEach(listener -> listener.onJobStateChanged(job, job.getState()))
        );

        return jobExecutorService.submit(() -> {
            try {
                V result = job.call();

                listeners.forEach(listener -> listener.onJobFinished(job));

                return result;
            } catch (Exception ex) {
                LOG.error("Exception while executing {}: {}", job.getName(), ex.getMessage(), ex);
                throw ex;
            }
        });

    }

    @Nonnull
    public Experiment getExperiment() {
        return experiment;
    }

    public Window getParentWindow() {
        return parentWindow;
    }

    public void setParentWindow(Window parentWindow) {
        this.parentWindow = parentWindow;
    }

    public void addListener(ExperimentControllerListener listener) {
        this.listeners.add(listener);
    }

    public void removeListener(ExperimentControllerListener listener) {
        this.listeners.remove(listener);
    }

    public ExperimentTester getExperimentTester() {
        return experimentTester;
    }

    public void setExperimentTester(ExperimentTester experimentTester) {
        this.experimentTester = experimentTester;
    }

    public <T extends ExperimentEvent> void addEventHandler(
            ExperimentEventType<T> type, ExperimentEventHandler<T> eventHandler) {
        eventHandlerManager.addExperimentEventHandler(type, eventHandler);
    }
}
